/* 
parser.h
Igor Terentyev.
********************************************************************************
Decomposes file/string into pairs of strings: [name][Value].

Separator control symbol between Name and Value is '='
(can be redefined in utils.h file).

Delimiters between Names/Values are any non-graphical symbols.

Names CANNOT be empty: if name not found between separators,
separator is ignored (treated as delimiter).

Values can be empty.

Quoute control symbol is '"' (can be redefined in utils.h file).
Everything between quotes is treated as graphical symbol.

'"' symbol (non-control) is given by [""].

Strings do not have to be null-terminated. String size is provided.
For just in case strings produced by parser are by one symbol longer
then their ireal sizes and have null-terminator in the extra space.

EXAMPLES:

[a=b]
==> [a][b].
--------------------------------------------------------------------------------

[a = b]
==> [a][b].
--------------------------------------------------------------------------------

[a =b = c = d]
==> [a][b],[c][d]; second '=' ignored since no name to the left from it.
--------------------------------------------------------------------------------

["  name w/ spaces" = "value w/ spaces and ""quotes"""]
==> [  name w/ spaces][value w/ spaces and "quotes"]
--------------------------------------------------------------------------------

[eq_sign="="]
==> [eq_sign][=].
--------------------------------------------------------------------------------

[a = 1 comment 1 b = 2 comment 2 c = 3]
==> [a][1],[b][2],[c][3].
--------------------------------------------------------------------------------

[ a

=

1]
==> [a][1].
--------------------------------------------------------------------------------

[a= = (empty parameter) b = 1 (non-empty parameter)]
==> [a][],[b][1].
*/
/*============================================================================*/

#ifndef __PARSER_H_
#define __PARSER_H_
/*----------------------------------------------------------------------------*/

#include "utils.h"
/*----------------------------------------------------------------------------*/
/*
Parameter structure.

int npar        :  number of parameters in the array.
SIZEDSTRING *ns :  names array.
SIZEDSTRING *vs :  values array.
char *buffer    :  allocated buffer.
*/
typedef struct
{
  int npar;
  SIZEDSTRING *ns;
  SIZEDSTRING *vs;
  char *buffer;
	int buffersize;
} PARARRAY;
/*----------------------------------------------------------------------------*/
/* 
Set parameter array (all fields) to zeros. 
*/
int ps_setnull(PARARRAY *parr);
/*----------------------------------------------------------------------------*/
/* 
Creates parameter array (STORAGE ALLOCATION) from file, argument list, string. 
*/
int ps_createfile(PARARRAY *parr, const char *filename);
int ps_createargs(PARARRAY *parr, int argc, char **argv);
int ps_createstrz(PARARRAY *parr, SIZEDSTRING str);
/*----------------------------------------------------------------------------*/
/* 
Destroy parameter array (STORAGE DEALLOCATION). 
*/
int ps_destroy(PARARRAY *parr);
/*----------------------------------------------------------------------------*/
/* 
Concatenate parameter arrays (STORAGE REALLOCATION).
"parr = parr + parr2".
*/
int ps_concat(PARARRAY *parr, PARARRAY parr2);
/*----------------------------------------------------------------------------*/
/* 
Get first occurence of data in the first occurence of the name. 

PARARRAY parr :  parameter array.
char *name    :  valiable name (null-terminated string).
type *p       :  variable pointer.

int return    :  error code.
*/
int ps_ffcstring(PARARRAY parr, const char *name, char **p);/* ALLOCATES MEMORY */
int ps_ffchar(PARARRAY parr, const char *name, char *p);
int ps_ffshort(PARARRAY parr, const char *name, short *p);
int ps_ffint(PARARRAY parr, const char *name, int *p);
int ps_fflong(PARARRAY parr, const char *name, long *p);
int ps_ffushort(PARARRAY parr, const char *name, unsigned short *p);
int ps_ffuint(PARARRAY parr, const char *name, unsigned int *p);
int ps_ffulong(PARARRAY parr, const char *name, unsigned long *p);
int ps_fffloat(PARARRAY parr, const char *name, float *p);
int ps_ffdouble(PARARRAY parr, const char *name, double *p);
int ps_ffreal(PARARRAY parr, const char *name, ireal *p);
/*----------------------------------------------------------------------------*/
/* 
Get first occurence of data in the last occurence of the name. 

PARARRAY parr :  parameter array.
char *name    :  valiable name (null-terminated string).
type *p       :  variable pointer.

int return    :  error code.
*/
int ps_flcstring(PARARRAY parr, const char *name, char **p);/* ALLOCATES MEMORY */
int ps_flchar(PARARRAY parr, const char *name, char *p);
int ps_flshort(PARARRAY parr, const char *name, short *p);
int ps_flint(PARARRAY parr, const char *name, int *p);
int ps_fllong(PARARRAY parr, const char *name, long *p);
int ps_flushort(PARARRAY parr, const char *name, unsigned short *p);
int ps_fluint(PARARRAY parr, const char *name, unsigned int *p);
int ps_flulong(PARARRAY parr, const char *name, unsigned long *p);
int ps_flfloat(PARARRAY parr, const char *name, float *p);
int ps_fldouble(PARARRAY parr, const char *name, double *p);
int ps_flreal(PARARRAY parr, const char *name, ireal *p);
/*----------------------------------------------------------------------------*/
/* 
Return number of parameter name occurences. 
*/
int ps_countnamez(PARARRAY parr, SIZEDSTRING name);
int ps_countname(PARARRAY parr, const char *name);
/*----------------------------------------------------------------------------*/
/* 
Return number of parameter type value occurences in the parameter value. 

int occ :  name occurence number.
*/
int ps_counttypez(PARARRAY parr, SIZEDSTRING name, int occ, int type);
int ps_counttype(PARARRAY parr, const char *name, int occ, int type);
/*----------------------------------------------------------------------------*/
/*
Extract type value.

int occ :  name occurence number.
int n   :  type occurence number in the parameter value.
*/
int ps_gettypez(PARARRAY parr, SIZEDSTRING name, int occ, int type, int n, void *p);
int ps_gettype(PARARRAY parr, const char *name, int occ, int type, int n, void *p);
/*----------------------------------------------------------------------------*/
/* 
Returns sized string value of the parameter from name.

int occ :  name occurence number.
*/
int ps_getvalz(PARARRAY parr, SIZEDSTRING name, int occ, SIZEDSTRING *val);
int ps_getval(PARARRAY parr, const char *name, int occ, SIZEDSTRING *val);
/*----------------------------------------------------------------------------*/
/* 
Return index of parameter occurence. 
Returns -1 if the occurence not found.

int occ :  name occurence number.
*/
int ps_nameindexz(PARARRAY parr, SIZEDSTRING name, int occ);
int ps_nameindex(PARARRAY parr, const char *name, int occ);
/*----------------------------------------------------------------------------*/
/*
Convert zero terminated string to sized string (NO ALLOCATION).
*/
SIZEDSTRING ps_name2z(const char *name);
/*----------------------------------------------------------------------------*/
/*
Seeks next value and converts it.
MAY ALLOCATE (if type is cstring).
*/
int ps_val2type(SIZEDSTRING val, int type, void *p, SIZEDSTRING *valend);
/*----------------------------------------------------------------------------*/
/*
Output all parameters.
*/
int ps_printall(PARARRAY parr, FILE *stream);
/*----------------------------------------------------------------------------*/
/*
Full copy function
*/
int ps_copy(PARARRAY rhs, PARARRAY *lhs);

#endif /*__PARSER_H_*/
