#include "wavefun.h"

#define VCONST 1.5
#define CFL 0.7

int main(int argc, char ** argv) {

  /* BEGIN DECLARATIONS */

  /* input data */

  WINFO ri;              /* struct for command line input */	
  FILE * fpt;            /* trace stream */
  FILE * fpm;            /* movie stream */
  FILE * fpv;            /* velocity file */
  FILE * fprm;           /* receiver movie stream */
  FILE * fpim;           /* image file */
 
  /* workspace */

  float * restrict v;    /* velocity field */
  float * restrict p1;   /* pressure field, current time step */
  float * restrict p0;   /* pressure field, last time step */
  float * restrict rp1;  /* receiver pressure field, current time step */
  float * restrict rp0;  /* receiver pressure field, last time step */
  float * tr;            /* traces */
  float * tmp;           /* swap */
  float * restrict imag; /* image */ 
  
  int ix, iz, it;        /* space, time counters */
  int isrc;              /* source counter */
  int isx;               /* source location, in units of dx */
  int imf;               /* movie frame counter */
  int nxz;               /* number of spatial grid points */
  int nz;                /* local number of gridpoints in z */
  int ih;                /* offset counter, from -ihmax */
  int ihind;             /* offset counter, from 0 */
  int nhz;               /* number of points in each image panel */
  int imid;              /* midpoint counter */
  int nmid;              /* number of midpoints */
  int ioff;              /* array offset in space */
  int ihoff;             /* array offset in image */
  int ntr;               /* number of traces */
  int nsam;              /* number of samples in shot record */
  float rz, rx, s;       /* precomputed coefficients */
  float vmax, vmin;      /* max and min velocity */
  float two;             /* two */
  
  /* END DECLARATIONS */

  if (getinputs(argc,argv,&ri)) {
    fprintf(stderr,"ERROR: failed to read legit inputs from command line\n");
    exit(1);
  }

  /* compute number of spatial grid points */
  nxz=ri.nx * ri.nz; 

  /* compute number of traces, samples in each record */
  ntr=ri.igxend-ri.igxbeg+1;
  nsam=ntr*ri.nt;

  /* compute number of samples in each image panel */
  nhz=ri.nz*(2*ri.ihmax+1);

  /* compute number of midpoints */
  nmid=1+((ri.imend-ri.imbeg)/(ri.imskip));

  /* allocate, initialize p0, p1, v, traces, rp0, rp1 */
  if (!(p0=(float*)malloc(nxz*sizeof(float))) ||
      !(p1=(float*)malloc(nxz*sizeof(float))) ||
      !(rp0=(float*)malloc(nxz*sizeof(float))) ||
      !(rp1=(float*)malloc(nxz*sizeof(float))) ||
      !(imag=(float*)malloc(nhz*nmid*sizeof(float))) ||
      !(v=(float*)malloc(nxz*sizeof(float))) ||
      !(tr=(float*)malloc(nsam*sizeof(float)))) {
    fprintf(stderr,"ERROR:: failed to allocate memory for migration comp\n");
    exit(1);
  }

  /* initialize pressure fields, image */
  fzeros(p0,nxz);
  fzeros(p1,nxz);
  fzeros(imag,nhz*nmid);
    
  /* assign constant velocity... */
  if (!ri.vfile){
    fassign(v,VCONST,nxz);
    fprintf(stderr,"velocity is constant\n");
  }
  /* else read from file */
  else {
    fpv=NULL;
    fpv=fopen(ri.vfile,"r");
    if (!fpv || (nxz != fread(v,sizeof(float),nxz,fpv))) {
      fprintf(stderr,"ERROR; failed to open velocity file %s, or to read %d words\n",
              ri.vfile,nxz);
      exit(1);
    }
    fclose(fpv);
  }

  /* open trace file for reading */
  if (ri.tfile){
    fpt=NULL;
    fpt=fopen(ri.tfile,"r");
    if (!fpt) {
      fprintf(stderr, "ERROR: failed to open trace file %s\n",ri.tfile);  
      exit(1);
    }
  }
  else {
    fprintf(stderr,"ERROR: no input trace file\n");
    exit(1);
  }

  /* open source wavefield movie file for reading */
  fpm=NULL;
  if (ri.mfile) {
    fpm=fopen(ri.mfile,"r");
    if (!fpm) {
      fprintf(stderr,"ERROR: failed to open source wavefield movie file %s\n",
	      ri.mfile);
      exit(1);
    }
  } 
  
  /* open image file for writing */
  fpim=NULL;
  if (ri.imfile)
    fpim=fopen(ri.imfile,"w");
   
  /* open receiver movie file for writing */
  fprm=NULL;
  if (ri.rmfile)
    fprm=fopen(ri.rmfile,"w");
  
  /* CFL, sanity checks */
  vmax=fgetmax(v,nxz);
  vmin=fgetmin(v,nxz);
  if (vmax*ri.dt>CFL*fmaxf(ri.dx,ri.dz)) {
    fprintf(stderr,"ERROR: CFL criterion violated\n");
    fprintf(stderr,"vmax=%e dx=%e dz=%e dt=%e\n",vmax,ri.dx,ri.dz,ri.dt);
    fprintf(stderr,"max permitted dt=%e\n",CFL*fmaxf(ri.dx,ri.dz)/vmax);
    exit(1);
  }
  if (vmin<=0.0) {
    fprintf(stderr,"ERROR: min velocity nonpositive\n");
    exit(1);
  }

  /* only square of velocity array needed from here on */
  fsquare(v,nxz);

  /* precalculate some coefficients */
  rz=ri.dt*ri.dt/(ri.dz*ri.dz);
  rx=ri.dt*ri.dt/(ri.dx*ri.dx);
  s =2.0*(rz+rx);
  two=2.0;
  nz=ri.nz;

  /* source loop */
  isrc=0;
  isx=ri.isxbeg;

  while (isx <= ri.isxend) { 
  
    /* FORWARD TIME LOOP */

    /* seek to begin of source wavefield file */
    if (fpm) fseek(fpm,0L,SEEK_SET);

    /* compute the source wavefile for one shot */
    fzeros(p0,nxz);
    fzeros(p1,nxz);
 
    /* precondition: before step it, p0 = time (it-1)*dt,
       p1 = time it*dt.
       postcondition: after step it, p0 = time it*dt, 
       p1 = time (it+1)*dt
    */

    for(it=0;it<ri.nt;it++) {

      /* construct next time step, overwrite on p0 */
      for (ix=1;ix<ri.nx-1;ix++) {
        for (iz=1;iz<nz-1;iz++) {
          ioff=iz+ix*nz;
          p0[ioff]=two*p1[ioff]-p0[ioff]
            +v[ioff]*(rz*(p1[ioff+1]+p1[ioff-1]) +
                      rx*(p1[ioff+nz]+p1[ioff-nz]) -
                      s*p1[ioff]);

        }
      }
 
      p0[ri.isz+isx*nz]+=fgetrick(it*ri.dt,ri.freq);

      /* swap pointers */
      tmp=p0;
      p0=p1;
      p1=tmp;

      /* optionally write source wavefield at time it to file -
         note that after pointer swap this is now p0! 
      */
      if (fpm) fwrite(p0,sizeof(float),nxz,fpm);
    }

    /* REVERSE TIME LOOP */
        
    fzeros(rp0,nxz);
    fzeros(rp1,nxz);
    imf = 0;
  
    /* compute receiver wavefield backward in time and image by
       crosscorrelation at zero time, space lag */
	
    /* read traces for this shot */
    if (fseek(fpt,nsam*isrc*sizeof(float),SEEK_SET)) {
      fprintf(stderr,"ERROR: seek failed in trace file %s, shot record %d\n",
	      ri.tfile,isrc);
      fprintf(stderr,"samples per shot record = %d\n",nsam);
      exit(1);
    }
    if (ri.nt*ntr != fread(tr,sizeof(float),nsam,fpt)) {
      fprintf(stderr,"ERROR: read failed in trace file %s, shot record %d\n",
	      ri.tfile,isrc);
      fprintf(stderr,"samples per shot record = %d\n",nsam);
      exit(1);
    }

    /* if source wavefields are to be time-stepped backwards,
       swap pointers so that p0 = time nt*dt 
    */
    if (!fpm) {
      tmp=p0;
      p0=p1;
      p1=tmp;
    }
       
    for (it=ri.nt-1;it>-1;it--) {
		
      /* construct next time step (backward) of receiver wavefield,
	 overwrite on rp0.  do same to source wavefield if not saved
	 to disk

	 precondition: before step it, rp0 = time (it+1)*dt,
	 rp1 = time it*dt.
	 postcondition: after step it, rp0 = time it*dt, 
	 rp1 = time (it-1)*dt
	 same for p0, p1
      */       
      if (fpm) {
	for (ix=1;ix<ri.nx-1;ix++) {
	  for (iz=1;iz<nz-1;iz++) {
	    ioff=iz+ix*nz;
	    rp0[ioff]=two*rp1[ioff]-rp0[ioff]
	      +v[ioff]*(rz*(rp1[ioff+1]+rp1[ioff-1]) +
			rx*(rp1[ioff+nz]+rp1[ioff-nz]) -
			s*rp1[ioff]);
	  }
	}
	/* swap pointers */
	tmp=rp0;
	rp0=rp1;
	rp1=tmp;
		
      }
      else {

	/* going backwards in time, SUBTRACT source BEFORE step */
	p0[ri.isz+isx*nz]-=fgetrick(it*ri.dt,ri.freq);

	for (ix=1;ix<ri.nx-1;ix++) {
	  for (iz=1;iz<nz-1;iz++) {
	    ioff=iz+ix*nz;
	    rp0[ioff]=two*rp1[ioff]-rp0[ioff]
	      +v[ioff]*(rz*(rp1[ioff+1]+rp1[ioff-1]) +
			rx*(rp1[ioff+nz]+rp1[ioff-nz]) -
			s*rp1[ioff]);
	  }
	  for (iz=1;iz<nz-1;iz++) {
	    ioff=iz+ix*nz;
	    p0[ioff]=two*p1[ioff]-p0[ioff]
	      +v[ioff]*(rz*(p1[ioff+1]+p1[ioff-1]) +
			rx*(p1[ioff+nz]+p1[ioff-nz]) -
			s*p1[ioff]);
	  }
	}
	/* swap pointers */
	tmp=rp0;
	rp0=rp1;
	rp1=tmp;
	tmp=p0;
	p0=p1;
	p1=tmp;
      }

      /* traces act as sources for backwards propagation of receiver field */
      for (ix=0;ix<ntr;ix++) 
	rp1[ri.igz+(ri.igxbeg+ix)*nz]+=tr[ix*ri.nt+it];
		
      /* option: read source field from movie file - seek is necessary
         because the part of the file for each shot record is read
         BACKWARDS */
      if (fpm) {
	if (fseek(fpm,(nxz*it)*sizeof(float),SEEK_SET)) {
	  fprintf(stderr,"ERROR: seek failed, source wavefield file=%s word=%d\n",
		  ri.mfile,nxz*ri.nt*isrc+nxz*it);
	  fprintf(stderr,"source wavefield of size %d at time step %d nt=%d\n",
		  nxz,it,ri.nt);
	  fprintf(stderr,"shot number = %d shot coord (units of dx) = %d\n",
		  isrc,isx);
	  exit(1);
	}
	if (nxz!=fread(p0,sizeof(float),nxz,fpm)) {
	  fprintf(stderr,"ERROR: failed to read in source wavefield\n",
		  nxz*ri.nt*isrc+nxz*it);
	  fprintf(stderr,"of size %d at time step %d nt=%d\n",
		  nxz,it,ri.nt);
	  fprintf(stderr,"shot number = %d shot coord (units of dx) = %d\n",
		  isrc,isx);
	  exit(1);
	}
      }

      /* imaging condition - loop over midpoints */
      for (imid=0;imid<nmid;imid++) {
	/* compute midpoint in units of dz */
	ix=ri.imbeg+imid*ri.imskip;
	/* loop over offset */
	for (ih=-ri.ihmax;ih<ri.ihmax+1;ih++) {
	  ihind=ih+ri.ihmax;
	  /* loop over depth */
	  for (iz=1;iz<nz-1;iz++) {
	    /* array offset of space column under midpoint */
	    ioff =iz+ix*nz;
	    /* array offset of image column */
	    ihoff=iz+ihind*nz+imid*nhz;
	    imag[ihoff] 
	      += p0[ioff-ih*nz]*rp1[ioff+ih*nz];
	  }
	}
      }
        
      /* write receiver movie snap to file if necessary */
      if (fprm) {
	fwrite(rp1,sizeof(float),nxz,fprm);
	imf++;
      }
      
      /* next t */
      
    }
    
    isx += ri.iskip;
    isrc++;
    
    /* next shot record */
  }
  
  /* write image file */
  if (fpim) fwrite(imag,sizeof(float),nhz*nmid,fpim);
  
  /* clean up */
  fclose(fpm);
  fclose(fpt);
  fclose(fpim);
  fclose(fprm);
  free(v);
  free(p0);
  free(p1);
  free(tr);
  free(rp0);
  free(rp1);
  free(imag);
  
  /* ta-ta */
  
  if (ri.rmfile)
    printf("receiver wavefield: %d shots, %d frames, size of %d x %d, written to %s\n",
	   isrc,imf,ri.nx,nz,ri.rmfile);
  if (ri.imfile)
    printf("image: size %d (nz) x %d (nh) x %d (nmid), written to %s\n",
	   nz,2*ri.ihmax+1,nmid,ri.imfile);
  
  exit(0);
}    
