#include "wavefun.h"
#include "mpi.h"

/* constant velocity constant */
#define VCONST 1.5

/* verbosity flag */
/*
#define VERB
*/

/* cfl number appropriate for scheme */
#define CFL 0.7

extern void mpi_step_forward(float *, 
			     float *,
			     float *,
			     int,
			     int,
			     float,
			     float,
			     float,
			     int,
			     int,
			     MPI_Status *);

int main(int argc, char ** argv) {

  /* BEGIN DECLARATIONS */

  WINFO wi;        /* struct for command line input */
  FILE * fpt;      /* trace stream */
  /* NO MOVIES */
  /* FILE * fpm; */     /* movie stream */
  FILE * fpv;      /* velocity file */

  /* workspace */

  float * restrict v;       /* velocity field */
  float * restrict p1;      /* pressure field, current time step */
  float * restrict p0;      /* pressure field, last time step */

  float * tr;      /* storage for traces */
  float * tmp;     /* used to swap p1 and p0 */

  int ix, it;      /* counters */
  int isrc;        /* source counter */
  /* NO MOVIES */
  /* int imf; */         /* movie frame counter */
  int isx;         /* source location, in units of dx */
  int nxz;         /* number of spatial grid points */
  int nz;          /* number of depth points */
  int ntr;         /* number of traces */
  int nsam;        /* number of trace samples */
  float rz,rx,s;   /* precomputed coefficients */
  float vmax,vmin; /* max, min velocity values */
  float two;       /* two */

  int size;        /* MPI: number of blocks */
  int rank;        /* MPI: block index = rank, procid */
  int nxblk;       /* MPI: number of x indices, this block (w/ ghost cells) */
  int nxzblk;      /* MPI: number of points, this block */
  int ixbegblk;    /* MPI: left x index, this block (w/ ghost cells) */
  int ixendblk;    /* MPI: right x index, this block (w/ ghost cells) */
  int ntrblk;      /* MPI: number of traces in this block */
  int nsamblk;     /* MPI: number of trace samples in this block */
  int igxbegblk;   /* MPI: left trace index, this block */
  int igxendblk;   /* MPI: right trace index, this block */
  MPI_Status stat; /* MPI status workspace */

  /* END DECLARATIONS */

  /* create MPI communicators */
  MPI_Init(&argc,&argv);

  /* number of grid blocks, index of this one */
  MPI_Comm_size(MPI_COMM_WORLD,&size);
  MPI_Comm_rank(MPI_COMM_WORLD,&rank);
    
  /* get arguments */
  if (getinputs(argc,argv,&wi)) {
    fprintf(stderr,"ERROR, rank=%d: failed to read legit inputs from command line\n",
	    rank);
    MPI_Finalize(); exit(1);
  }

  /* calculate x-extent of this grid block, updated part */
  nxblk=wi.nx/size;
  /* if remainder then up by one */
  if (nxblk*size < wi.nx) nxblk++;

  /* compute beginning and ending block indices - note that first and 
     last cols are ghost cells, except on boundary */
  /* updated block begin */
  ixbegblk = nxblk*rank;
  /* add left ghost cells */
  if (rank>0) ixbegblk-=1;
  /* updated block end */
  ixendblk = iwave_min(nxblk*(rank+1)-1,wi.nx-1);
  /* add right ghost cells */
  if (rank<size-1) ixendblk+=1;
  /* update nxblk to include ghost cells */
  nxblk=ixendblk-ixbegblk+1;

  /* compute number of spatial grid points */
  nxz=wi.nx * wi.nz;

  /* compute number of spatial grid points in this block */
  nxzblk=nxblk*wi.nz;

  /* compute number of traces, samples in each record */
  ntr=wi.igxend-wi.igxbeg+1;
  nsam=ntr*wi.nt;

  /* compute left and right trace indices, this block, not
     counting ghost cells */
  if (rank==0) {
    igxbegblk=iwave_max(wi.igxbeg,ixbegblk);
  }
  else {
    igxbegblk=iwave_max(wi.igxbeg,ixbegblk+1);
  }
  if (rank==size-1) {
    igxendblk=iwave_min(wi.igxend,ixendblk);
  }
  else {
    igxendblk=iwave_min(wi.igxend,ixendblk-1);
  }
  /* compute number of traces, samples in each record - note that it may 
     be zero if the cable does not intersect the block */
  ntrblk=iwave_max(0,igxendblk-igxbegblk+1);
  nsamblk=ntrblk*wi.nt;

#ifdef VERB
  fprintf(stderr,"\nblock parameters for rank = %d: \nleft index limit = %d\nright index limit= %d \nnumber of cols in block = %d\nleft trace indexig = %d\nright trace index = %d\nnumber of traces in block = %d\nnumber of samples in block = %d\n",rank,ixbegblk,ixendblk,nxblk,igxbegblk,igxendblk,ntrblk,nsamblk);
#endif

  /* allocate, initialize p0, p1, v, traces 
     NB: for parallel case, force allocation of at least one trace
     as workspace on p0, used to populate output file */
  if (!(p0=(float*)malloc(nxzblk*sizeof(float))) ||
      !(p1=(float*)malloc(nxzblk*sizeof(float))) ||
      !(v =(float*)malloc(nxzblk*sizeof(float))) ||
      !(tr=(float*)malloc(iwave_max(nsamblk,wi.nt)*sizeof(float)))) {
    fprintf(stderr,"ERROR, rank=%d: failed to allocate memory for wave comp\n",
	    rank);
    MPI_Finalize(); exit(1);
  }

  /* assign velocity: either constant */
  if (!wi.vfile) 
    fassign(v,VCONST,nxzblk);
  /* else read from file */
  else {
    fpv=NULL;
    fpv=fopen(wi.vfile,"r");
    if (!fpv) {
      fprintf(stderr,"ERROR, rank=%d: failed to open velocity file %s\n",
	      rank,wi.vfile);
      MPI_Finalize(); exit(1);
    }
    /* seek to file segment corresponding to this block */
    if (fseek(fpv,ixbegblk*wi.nz*sizeof(float),SEEK_SET)) {
      fprintf(stderr,"ERROR, rank=%d: failed to seek to word %d in velocity file\n",
	      rank,ixbegblk*wi.nz);
      MPI_Finalize(); exit(1);
    }
    /* read part of velocity corresponding to this block - nxblk columns */
    if (nxzblk != fread(v,sizeof(float),nxzblk,fpv)) {
      fprintf(stderr,"ERROR, rank=%d; failed to read %d words from velocity file\n",
	      rank,nxzblk);
      MPI_Finalize(); exit(1);
    }
    fclose(fpv);
  }
  
  /* CFL, sanity checks */
  vmax=fgetmax(v,nxzblk);
  vmin=fgetmin(v,nxzblk);
  if (vmax*wi.dt>CFL*fmaxf(wi.dx,wi.dz)) {
    fprintf(stderr,"ERROR, rank=%d: CFL criterion violated\n",rank);
    fprintf(stderr,"vmax=%e dx=%e dz=%e dt=%e\n",vmax,wi.dx,wi.dz,wi.dt);
    fprintf(stderr,"max permitted dt=%e\n",CFL*fmaxf(wi.dx,wi.dz)/vmax);
    MPI_Finalize(); exit(1);
  }
  if (vmin<=0.0) {
    fprintf(stderr,"ERROR, rank=%d: min velocity nonpositive\n",rank);
    MPI_Finalize(); exit(1);
  }

  /* only square of velocity array needed from here on */
  fsquare(v,nxzblk);

  /* precalculate some coefficients */
  rz=wi.dt*wi.dt/(wi.dz*wi.dz);
  rx=wi.dt*wi.dt/(wi.dx*wi.dx);
  s =2.0*(rz+rx);
  two=2.0;
  nz=wi.nz;

  /* open trace file if necessary */
  fpt=NULL;
  if (wi.tfile) {
    if (!(fpt=fopen(wi.tfile,"w"))) {
      fprintf(stderr,"ERROR, rank=%d: failed to open trace file %s for write\n",
	      rank,wi.tfile);
      MPI_Finalize(); exit(1);
    }
    /* populate trace file so that seek will succeed - only on p0 */
    if (rank==0) {
#ifdef VERB
      fprintf(stderr,"initialize %d traces\n",ntr);
#endif
      fzeros(tr,wi.nt);
      for (ix=0;ix<ntr;ix++) {
	fwrite(tr,sizeof(float),wi.nt,fpt);
      }
    }
  }

  /* open movie file if necessary */
  /* NB FOR NOW NO MOVIES IN PARALLEL
  fpm=NULL;
  if (wi.mfile)
    fpm=fopen(wi.mfile,"w");
  */

#ifdef VERB
  fprintf(stderr,"rank=%d - start time loop\n",rank);
#endif

  /* shot loop */
  isrc=0;
  isx=wi.isxbeg;
  while (isx <= wi.isxend) {

    /* initialize pressure fields, traces */
    fzeros(p0,nxzblk);
    fzeros(p1,nxzblk);
    fzeros(tr,nsamblk);
    
    /* initialize movie frame counter */
    /*    imf=0;*/

    /* time loop */
    for (it=0;it<wi.nt;it++) {
      
      /* construct next time step, overwrite on p0 */

      mpi_step_forward(p0,p1,v,wi.nz,nxblk,rz,rx,s,rank,size,&stat);

      /* tack on source - if in this block! Note that x index of 
	 source is relativized to block to compute offset
	 Note that the admit condition is simplified for (2,2)
	 scheme, as the updated points start at ixbegblk+1 and
	 end at ixendblk-1, for all blocks.
      */
      if (isx>ixbegblk && isx < ixendblk)
	p0[wi.isz+(isx-ixbegblk)*wi.nz]+=fgetrick(it*wi.dt,wi.freq);
      
      /* swap pointers */
      tmp=p0;
      p0=p1;
      p1=tmp;
      
      /* store trace samples if necessary */
      if (wi.tfile) 
	for (ix=0;ix<ntrblk;ix++) 
	  tr[ix*wi.nt+it]=p1[(igxbegblk-ixbegblk+ix)*wi.nz+wi.igz];
      
      /* write movie snap to file if necessary */
      /* NO MOVIES
      if (fpm && wi.nm && !(it%wi.nm)) {
	fwrite(p1,sizeof(float),nxz,fpm);
	imf++;
      }
      */
      
      /* next t */
    }

#ifdef VERB
    fprintf(stderr,"rank=%d write traces\n",rank);
#endif

    /* write traces to file if necessary */
    if (fpt) {
#ifdef VERB
      fprintf(stderr,"write %d samples = %d traces at offset %d\n",
	      nsamblk,ntrblk,(igxbegblk-wi.igxbeg)*wi.nt*sizeof(float));
#endif
      if (fseek(fpt,(igxbegblk-wi.igxbeg)*wi.nt*sizeof(float),SEEK_SET)) {
	fprintf(stderr,"ERROR, rank=%d: failed to seek to word %d in trace file\n",
		rank,igxbegblk*wi.nt);
	MPI_Finalize(); exit(1);
      }
      fwrite(tr,sizeof(float),nsamblk,fpt);
    }

    isx += wi.iskip;
    isrc++;
  } 

#ifdef VERB
  fprintf(stderr,"rank=%d clean up\n",rank);
#endif

  /* clean up */
  if (fpt) fclose(fpt);

  /*  if (fpm) fclose(fpm);*/

  /* ta-ta from P0 */
  if (wi.tfile && rank==0) 
    fprintf(stderr,"wave, rank=%d: %d shots, each with %d traces, each with %d samples, written to %s\n",
	   rank,isrc,ntr,wi.nt,wi.tfile);
  /*
  if (wi.mfile) 
    printf("wave, rank=%d: %d movies, each with %d frames of size %d x %d, written to %s\n",
	   rank,isrc,imf,wi.nz,wi.nx,wi.mfile);
  */

  /* shut down MPI */
  MPI_Finalize();

  free(v);
  free(p0);
  free(p1);
  free(tr);

  exit(0);
}
