#include "wavefun.h"

#define VCONST 1.5
#define CFL 0.7

extern void step_forward(float *, 
			 float *,
			 float *,
			 int,
			 int,
			 float,
			 float,
			 float);

int main(int argc, char ** argv) {

  /* BEGIN DECLARATIONS */

  /* input data */

  WINFO ri;              /* struct for command line input */	
  FILE * fpt;            /* trace stream */
  FILE * fpv;            /* velocity file */
  FILE * fpim;           /* image file (optional) */
  FILE * fpm;            /* source wavefield movie (optional) */
  FILE * fprm;           /* receiver wavefield movie (optional) */

  /* workspace */

  float * restrict v;    /* velocity field */
  float * restrict p1;   /* pressure field, current time step */
  float * restrict p0;   /* pressure field, last time step */
  float * restrict rp1;  /* receiver pressure field, current time step */
  float * restrict rp0;  /* receiver pressure field, last time step */
  float * tr;            /* traces */
  float * tmp;           /* swap */
  float * imag;          /* image */ 

  int ix, iz, it;        /* counters */
  int isrc;              /* source counter */
  int isx;               /* source location, in units of dx */
  int imf;               /* movie frame counter */
  int nxz;               /* number of spatial grid points */
  int nz;                /* local storage for number of depth points */
  int ioff;              /* array offset */
  int ntr;               /* number of traces */
  int nsam;              /* number of samples in shot record */
  float rz, rx, s;       /* precomputed coefficients */
  float vmax, vmin;      /* max and min velocity */ 
  float two;             /* two */

  /* END DECLARATIONS */

  /* PARSE ARGUMENTS */

  if (getinputs(argc,argv,&ri)) {
    fprintf(stderr,"ERROR: failed to read legit inputs from command line\n");
    exit(1);
  }

  /* ALLOCATE WORKSPACE */

  /* compute number of spatial grid points */
  nxz=ri.nx * ri.nz;

  /* compute number of traces, samples in each record */
  ntr=ri.igxend-ri.igxbeg+1;
  nsam=ntr*ri.nt;

  /* allocate, initialize p0, p1, v, traces, rp0, rp1 */
  if (!(p0=(float*)malloc(nxz*sizeof(float))) ||
      !(p1=(float*)malloc(nxz*sizeof(float))) ||
      !(rp0=(float*)malloc(nxz*sizeof(float))) ||
      !(rp1=(float*)malloc(nxz*sizeof(float))) ||
      !(imag=(float*)malloc(nxz*sizeof(float))) ||
      !(v=(float*)malloc(nxz*sizeof(float))) ||
      !(tr=(float*)malloc(nsam*sizeof(float)))) {
    fprintf(stderr,"ERROR:: failed to allocate memory for migration comp\n");
    exit(1);
  }

  /* INITIALIZE FILE I/O */

  /* required: initialize velocity - either assign constant... */
  if (!ri.vfile){
    fassign(v,VCONST,nxz);
    fprintf(stderr,"velocity is constant\n");
  }
  /* else read from file */
  else {
    fpv=NULL;
    fpv=fopen(ri.vfile,"r");
    if (!fpv || (nxz != fread(v,sizeof(float),nxz,fpv))) {
      fprintf(stderr,"ERROR; failed to open velocity file %s, or to read %d words\n",
              ri.vfile,nxz);
      exit(1);
    }
    fclose(fpv);
  }

  /* required: open trace file for reading */
  if (ri.tfile){
    fpt=NULL;
    fpt=fopen(ri.tfile,"r");
    if (!fpt) {
      fprintf(stderr, "ERROR: failed to open trace file %s\n",ri.tfile);  
      exit(1);
    }
  }
  else {
    fprintf(stderr,"ERROR: no input trace file\n");
    exit(1);
  }

  /* optional: open source wavefield movie file for reading and writing */
  fpm=NULL;
  if (ri.mfile) {
    fpm=fopen(ri.mfile,"w+");
    if (!fpm) {
      fprintf(stderr,"ERROR: failed to open source wavefield movie file %s\n",
	      ri.mfile);
      exit(1);
    }
  } 
  
  /* optional (but why do it otherwise): open image file for writing */
  fpim=NULL;
  if (ri.imfile) {
    fpim=fopen(ri.imfile,"w");
    if (!fpim) {
      fprintf(stderr,"ERROR: failed to open image file %s\n",
	      ri.imfile);
      exit(1);
    }
  }

  /* optional: open receiver movie file for writing */
  fprm=NULL;
  if (ri.rmfile) {
    fprm=fopen(ri.rmfile,"w");
    if (!fprm) {
      fprintf(stderr,"ERROR: failed to open receiver wavefield movie file %s\n",
	      ri.rmfile);
      exit(1);
    }
  }

  /* PARAMETERS */

  /* CFL, sanity checks */
  vmax=fgetmax(v,nxz);
  vmin=fgetmin(v,nxz);
  if (vmax*ri.dt>CFL*fmaxf(ri.dx,ri.dz)) {
    fprintf(stderr,"ERROR: CFL criterion violated\n");
    fprintf(stderr,"vmax=%e dx=%e dz=%e dt=%e\n",vmax,ri.dx,ri.dz,ri.dt);
    fprintf(stderr,"max permitted dt=%e\n",CFL*fmaxf(ri.dx,ri.dz)/vmax);
    exit(1);
  }
  if (vmin<=0.0) {
    fprintf(stderr,"ERROR: min velocity nonpositive\n");
    exit(1);
  }

  /* only square of velocity array needed from here on */
  fsquare(v,nxz);

  /* precalculate some coefficients */
  rz=ri.dt*ri.dt/(ri.dz*ri.dz);
  rx=ri.dt*ri.dt/(ri.dx*ri.dx);
  s =2.0*(rz+rx);
  two=2.0;
  nz=ri.nz;

  /* initialize image field */
  fzeros(imag,nxz);
    
  /* SOURCE LOOP */

  isrc=0;
  isx=ri.isxbeg;

  while (isx <= ri.isxend) { 

    /* FORWARD TIME LOOP */

    /* seek to begin of source wavefield file */
    if (fpm) fseek(fpm,0L,SEEK_SET);

    /* compute the source wavefile for one shot */
    fzeros(p0,nxz);
    fzeros(p1,nxz);
 
    /* precondition: before step it, p0 = time (it-1)*dt,
       p1 = time it*dt.
       postcondition: after step it, p0 = time it*dt, 
       p1 = time (it+1)*dt
    */

    for(it=0;it<ri.nt;it++) {

      /* construct next time step, overwrite on p0 */
      step_forward(p0,p1,v,ri.nz,ri.nx,rz,rx,s);
 
      p0[ri.isz+isx*nz]+=fgetrick(it*ri.dt,ri.freq);

      /* swap pointers */
      tmp=p0;
      p0=p1;
      p1=tmp;

      /* optionally write source wavefield at time it to file -
         note that after pointer swap this is now p0! 
      */
      if (fpm) fwrite(p0,sizeof(float),nxz,fpm);
    }

    /* REVERSE TIME LOOP */
        
    /* receiver wavefield initialization */
    fzeros(rp0,nxz);
    fzeros(rp1,nxz);
    imf = 0;
  
    /* compute receiver wavefield backward in time and image by
       crosscorrelation at zero time, space lag */
	
    /* read traces for this shot */
    if (fseek(fpt,nsam*isrc*sizeof(float),SEEK_SET)) {
      fprintf(stderr,"ERROR: seek failed in trace file %s, shot record %d\n",
	      ri.tfile,isrc);
      fprintf(stderr,"samples per shot record = %d\n",nsam);
      exit(1);
    }
    if (ri.nt*ntr != fread(tr,sizeof(float),nsam,fpt)) {
      fprintf(stderr,"ERROR: read failed in trace file %s, shot record %d\n",
	      ri.tfile,isrc);
      fprintf(stderr,"samples per shot record = %d\n",nsam);
      exit(1);
    }

    /* if source wavefields are to be time-stepped backwards,
       swap pointers so that p0 = time nt*dt 
    */
    if (!fpm) {
      tmp=p0;
      p0=p1;
      p1=tmp;
    }
       
    for (it=ri.nt-1;it>-1;it--) {
		
      /* construct next time step (backward) of receiver wavefield,
	 overwrite on rp0.  do same to source wavefield if not saved
	 to disk

	 precondition: before step it, rp0 = time (it+1)*dt,
	 rp1 = time it*dt.
	 postcondition: after step it, rp0 = time it*dt, 
	 rp1 = time (it-1)*dt
	 same for p0, p1
      */       
      if (fpm) {

	step_forward(rp0,rp1,v,ri.nz,ri.nx,rz,rx,s);

	/* swap pointers */
	tmp=rp0;
	rp0=rp1;
	rp1=tmp;
		
      }
      else {

	/* going backwards in time, SUBTRACT source BEFORE step */
	p0[ri.isz+isx*nz]-=fgetrick(it*ri.dt,ri.freq);

	step_forward(rp0,rp1,v,ri.nz,ri.nx,rz,rx,s);
	step_forward(p0,p1,v,ri.nz,ri.nx,rz,rx,s);

	/* swap pointers */
	tmp=rp0;
	rp0=rp1;
	rp1=tmp;
	tmp=p0;
	p0=p1;
	p1=tmp;
      }

      /* traces act as sources for backwards propagation of receiver field */
      for (ix=0;ix<ntr;ix++) 
	rp1[ri.igz+(ri.igxbeg+ix)*nz]+=tr[ix*ri.nt+it];
		
      /* option: read source field from movie file - seek is necessary
         because the part of the file for each shot record is read
         BACKWARDS */
      if (fpm) {
	if (fseek(fpm,(nxz*it)*sizeof(float),SEEK_SET)) {
	  fprintf(stderr,"ERROR: seek failed, source wavefield file=%s word=%d\n",
		  ri.mfile,nxz*ri.nt*isrc+nxz*it);
	  fprintf(stderr,"source wavefield of size %d at time step %d nt=%d\n",
		  nxz,it,ri.nt);
	  fprintf(stderr,"shot number = %d shot coord (units of dx) = %d\n",
		  isrc,isx);
	  exit(1);
	}
	if (nxz!=fread(p0,sizeof(float),nxz,fpm)) {
	  fprintf(stderr,"ERROR: failed to read in source wavefield\n",
		  nxz*ri.nt*isrc+nxz*it);
	  fprintf(stderr,"of size %d at time step %d nt=%d\n",
		  nxz,it,ri.nt);
	  fprintf(stderr,"shot number = %d shot coord (units of dx) = %d\n",
		  isrc,isx);
	  exit(1);
	}
      }

      /* imaging condition */
      for (ix=1;ix<ri.nx-1;ix++) {
	for (iz=1;iz<nz-1;iz++) {
	  ioff=iz+ix*nz;
	  imag[ioff] += p0[ioff]*rp1[ioff];
	}
      }
        
      /* write receiver movie snap to file if necessary */
      if (fprm) {
        fwrite(rp1,sizeof(float),nxz,fprm);
        imf++;
      }
        
      /* next t */
    }

    isx += ri.iskip;
    isrc++;
		
    /* next shot record */
  }

  /* write image file */
  if (fpim) fwrite(imag,sizeof(float),nxz,fpim);

  /* clean up */
  fclose(fpt);
  if (fpim) fclose(fpim);
  if (fpm) fclose(fpm);
  if (fprm) fclose(fprm);
  free(v);
  free(p0);
  free(p1);
  free(tr);
  free(rp0);
  free(rp1);
  free(imag);
  /* ta-ta */

  if (ri.rmfile)
    printf("receiver wavefield: %d shots, %d frames, size of %d x %d, written to %s\n",
           isrc,imf,ri.nx,nz,ri.rmfile);
  if (ri.imfile)
    printf("image: size %d x %d, written to %s\n",
           nz,ri.nx,ri.imfile);

  exit(0);
}     
