#include "wave_selfdoc.h"
#include "wavefun.h"

/* constant velocity constant */
#define VCONST 1.5

/* cfl number appropriate for scheme */
#define CFL 0.7

extern void step24_forward(float *, 
			   float *,
			   float *,
			   int,
			   int,
			   float,
			   float,
			   float);

int main(int argc, char ** argv) {

  /* BEGIN DECLARATIONS */

  WINFO wi;        /* struct for command line input */
  FILE * fpt;      /* trace stream */
  FILE * fpm;      /* movie stream */
  FILE * fpv;      /* velocity file */

  /* workspace */

  float * restrict v;       /* velocity field */
  float * restrict p1;      /* pressure field, current time step */
  float * restrict p0;      /* pressure field, last time step */

  float * tr;      /* storage for traces */
  float * tmp;     /* used to swap p1 and p0 */

  int ix, it;      /* counters */
  int isrc;        /* source counter */
  int imf;         /* movie frame counter */
  int isx;         /* source location, in units of dx */
  int nxz;         /* number of spatial grid points */
  int nz;          /* local number of gridpoints */
  int ntr;         /* number of traces */
  int nsam;        /* number of trace samples */
  int nsrc;        /* number of shots */
  float rz,rx,s;   /* precomputed coefficients */
  float vmax,vmin; /* max, min velocity values */
  float two;       /* two */

  /* END DECLARATIONS */

  /* self-doc */
  xargc=argc;
  xargv=argv;
  requestdoc(1);

  /* read inputs from command line */
  if (getinputs(argc,argv,&wi)) {
    fprintf(stderr,"ERROR: failed to read legit inputs from command line\n");
    exit(1);
  }
    
  /* compute number of shots */
  nsrc = (wi.isxend-wi.isxbeg)/(wi.iskip); nsrc++;

  /* compute number of spatial grid points */
  /* 2,4 FEATURE: nz->nz+2, nx->nx+2 */
  nxz=(wi.nx+2) * (wi.nz+2);

  /* compute number of traces, samples in each record */
  ntr=wi.igxend-wi.igxbeg+1;
  nsam=ntr*wi.nt;

  /* allocate, initialize p0, p1, v, traces */
  if (!(p0=(float*)malloc(nxz*sizeof(float))) ||
      !(p1=(float*)malloc(nxz*sizeof(float))) ||
      !(v =(float*)malloc(nxz*sizeof(float))) ||
      !(tr=(float*)malloc(nsam*sizeof(float)))) {
    fprintf(stderr,"ERROR: failed to allocate memory for wave comp\n");
    exit(1);
  }

  /* assign velocity: either constant */
  fzeros(v,nxz);
  if (!wi.vfile) 
    fassign(v,VCONST,nxz);
  /* else read from file */
  else {
    fpv=NULL;
    fpv=fopen(wi.vfile,"r");
    if (!fpv)
      fprintf(stderr,"ERROR: failed to open velocity file %s\n",
              wi.vfile);
    /* 2,4 FEATURE: nz->nz+2, etc: read interior from file, extend by const */
    /* read into interior in x */
    for (ix=1;ix<wi.nx+1;ix++) {
      /* read into interior of each column */
      fread(&(v[ix*(wi.nz+2)+1]),sizeof(float),wi.nz,fpv);
      /* extend by const */
      v[ix*(wi.nz+2)]=v[ix*(wi.nz+2)+1];
      v[(ix+1)*(wi.nz+2)-1]=v[(ix+1)*(wi.nz+2)-2];
    }
    /* extend to 1st and last cols */
    for (ix=0;ix<wi.nz+2;ix++) {
      v[ix]=v[ix+wi.nz+2];
      v[ix+(wi.nx+1)*(wi.nz+2)]=v[ix+(wi.nx)*(wi.nz+2)];
    }
    fclose(fpv);
  }
  
  /* CFL, sanity checks */
  vmax=fgetmax(v,nxz);
  vmin=fgetmin(v,nxz);
  if (vmax*wi.dt>CFL*fmaxf(wi.dx,wi.dz)) {
    fprintf(stderr,"ERROR: CFL criterion violated\n");
    fprintf(stderr,"vmax=%e dx=%e dz=%e dt=%e\n",vmax,wi.dx,wi.dz,wi.dt);
    fprintf(stderr,"max permitted dt=%e\n",CFL*fmaxf(wi.dx,wi.dz)/vmax);
    exit(1);
  }
  if (vmin<=0.0) {
    fprintf(stderr,"ERROR: min velocity nonpositive\n");
    exit(1);
  }

  /* only square of velocity array needed from here on */
  fsquare(v,nxz);

  /* precalculate some coefficients */
  rz=wi.dt*wi.dt/(wi.dz*wi.dz);
  rx=wi.dt*wi.dt/(wi.dx*wi.dx);
  s =2.0*(rz+rx);
  two=2.0;
  nz=wi.nz;

  /* open trace file if necessary */
  fpt=NULL;
  if (wi.tfile) 
    if (!(fpt=fopen(wi.tfile,"w"))) {
      fprintf(stderr,"ERROR: failed to open trace file %s\n",wi.tfile);
      exit(1);
    }

  /* open movie file if necessary - option only available for one shot */
  fpm=NULL;
  if (wi.mfile && nsrc==1)
    if (!(fpm=fopen(wi.mfile,"w"))) {
      fprintf(stderr,"ERROR: failed to open movie file %s\n",wi.mfile);
      exit(1);
    }


  /* shot loop */
  isrc=0;
  isx=wi.isxbeg;
  while (isx <= wi.isxend) {

    /* initialize pressure fields, traces */
    fzeros(p0,nxz);
    fzeros(p1,nxz);
    fzeros(tr,nsam);
    
    /* initialize movie frame counter */
    imf=0;

    /* time loop */
    for (it=0;it<wi.nt;it++) {
      
      /* construct next time step, overwrite on p0 */

      /* 2,4 FEATURE: nz->nz+2, nx->nx+2 */
      step24_forward(p0,p1,v,wi.nz+2,wi.nx+2,rz,rx,s);
      
      /* tack on source */
      /* 2,4 FEATURE: nz->nz+2, isz->isz+1, isx->isx+1 */
      p0[(wi.isz+1)+(isx+1)*(wi.nz+2)]+=fgetrick(it*wi.dt,wi.freq);
      
      /* swap pointers */
      tmp=p0;
      p0=p1;
      p1=tmp;
      
      /* store trace samples if necessary */
      if (wi.tfile) 
	for (ix=0;ix<ntr;ix++) 
	  /* 2,4 FEATURE: nz->nz+2, igz->igz+1 */
	  tr[ix*wi.nt+it]=p1[(wi.igxbeg+ix+1)*(wi.nz+2)+wi.igz+1];
      
      /* write movie snap to file if necessary */
      if (fpm && wi.nm && !(it%wi.nm)) {
	/* 2,4 FEATURE: write interior of each interior column */
	for (ix=1;ix<wi.nx+1;ix++) 
	  fwrite(&(p1[ix*(wi.nz+2)+1]),sizeof(float),wi.nz,fpm);
	imf++;
      }
      
      /* next t */
    }

    /* write traces to file if necessary */
    if (fpt) 
      fwrite(tr,sizeof(float),nsam,fpt);

    isx += wi.iskip;
    isrc++;
  } 

  /* clean up */
  if (fpt) fclose(fpt);
  if (fpm) fclose(fpm);

  free(v);
  free(p0);
  free(p1);
  free(tr);

  /* ta-ta */
  if (wi.tfile) 
    printf("wave: %d shots, each with %d traces, each with %d samples, written to %s\n",
	   nsrc,ntr,wi.nt,wi.tfile);
  if (wi.mfile && nsrc==1) 
    printf("wave: %d movies, each with %d frames of size %d x %d, written to %s\n",
	   isrc,imf,wi.nz,wi.nx,wi.mfile);

  exit(0);
}
